"""
擬似イベントを生成するモジュール

```python
from haniwers import mimic

# ジェネレーターとして使う例
for fe in mimic.fake_events_generator(n=100, seed=None, interval="random"):
    # FakeEventオブジェクトを処理する
    print(fe.timestamp)


# 擬似イベントのデータフレームを取得する例
fake_data = mimic.fake_events(100)
```
"""

from pydantic import BaseModel
import pendulum
import random
import pandas as pd
import time
from datetime import datetime


class FakeEvent(BaseModel):
    """擬似イベント用のクラス"""

    timestamp: datetime = pendulum.now()
    """タイムスタンプ。擬似イベントを生成した日時"""
    top: int = 0
    """topレイヤーの擬似ヒット。0 - 10のランダム値"""
    mid: int = 0
    """midレイヤーの擬似ヒット。0 - 10のランダム値"""
    btm: int = 0
    """btmレイヤーの擬似ヒット。0 - 10のランダム値"""
    adc: int = 0
    """topレイヤーにヒットがあったときのADC値。0 - 1023のランダム値"""
    tmp: float = 0.0
    """擬似イベントの気温。25度くらいのランダム値"""
    atm: float = 0.0
    """擬似イベントの気圧。1気圧（101800 Pa）くらいのランダム値"""
    hmd: float = 0.0
    """擬似イベントの湿度。50%くらいのランダム値"""
    seed: int | None = None
    """乱数シード。デフォルト値は`None`"""
    is_real: bool = False
    """データ種類のフラグ。擬似イベントの場合は常に`False`に設定"""

    def model_post_init(self, __context) -> None:
        """メンバー変数を初期化する"""
        random.seed(self.seed)
        self.timestamp = pendulum.now()
        self.top = random.randint(0, 10)
        self.mid = random.randint(0, 10)
        self.btm = random.randint(0, 10)
        self.tmp = random.gauss(mu=25.0, sigma=3) + random.gauss()
        self.atm = random.gauss(mu=101000, sigma=1000) + random.gauss()
        self.hmd = random.gauss(mu=50, sigma=15) + random.gauss()

        if self.top > 0:
            self.adc = random.randint(1, 1024)


def fake_events_generator(
    n: int, seed: int | None = None, interval: str | int = "random"
):
    """指定した回数のFakeEventを任意の間隔で生成する

    :Args:
        - n (int): 生成するFakeEventの数
        - seed (int | None) : 乱数シード。デフォルトは `None`
        - interval (int | str) : 生成する間隔。デフォルトは "random"

    :yield:
        - FakeEvent: 擬似イベントのオブジェクト

    """
    for _ in range(n):
        yield FakeEvent(seed=seed)
        if isinstance(interval, int):
            time.sleep(interval)
        elif interval in ("random"):
            r = random.randint(1, 5)
            time.sleep(r)


def fake_events(n: int) -> pd.DataFrame:
    """指定した回数のFakeEventをデータフレームに変換する

    内部で ``fake_events_generator(n, seed=None, interval=0)``を実行した結果を
    ``pd.DataFrame``に変換している


    :Args:
        - n (int): 生成するFakeEventの数

    :Returns:
        - data (pd.DataFrame): 擬似イベントのデータフレーム

    """
    events = []
    for fe in fake_events_generator(n, seed=None, interval=0):
        e = fe.model_dump()
        events.append(e)
    data = pd.DataFrame(events)
    return data



#! /usr/bin/env python3

from pathlib import Path
from typing import Dict, List

import pandas as pd
from deprecated import deprecated
from loguru import logger

from .config import RunData
from .dataset import load_files


@deprecated(version="0.3.0", reason="Use preprocess.resample_data.")
def resample_data(data: pd.DataFrame, interval: int) -> pd.DataFrame:
    """データフレームをリサンプリングする

    ``time`` をキーにして ``interval`` 間隔でリサンプリングして集計する。
    集計する内容は下記の通り。

    #. ``size`` : サンプリング数（＝ヒット数）
    #. ``sum`` : 上段 / 中段 / 下段 の合計値
    #. ``cumsum`` : 上段 / 中段 / 下段 の累積和
    #. ``mean`` : 平均値
    #. ``std`` : （不偏）標準偏差
    #. ``var`` : （不偏）分散

    Returns
    -------
    pd.DataFrame
        ``time`` をキーにして結合したデータフレーム
    """

    # インデックスを time に変更
    data.index = data["time"]

    # サンプリング間隔（秒数で指定）
    freq = f"{interval}s"

    # サンプリング数を求める
    renames: Dict[int, str] = {0: "hit"}  # type: ignore
    # fmt: off
    _size = (
        data.resample(freq)
        .size()
        .reset_index().rename(columns=renames)
    )
    # fmt: on

    # 合計を求める
    names = [
        "adc",
        "hit_top",
        "hit_mid",
        "hit_btm",
    ]
    # fmt: off
    _sum = (
        data.resample(freq)[names]
        .sum()
        .reset_index()
    )
    # fmt: on

    # 平均値を求める
    names = [
        "adc",
        "hit_top",
        "hit_mid",
        "hit_btm",
        "tmp",
        "atm",
        "hmd",
    ]
    renames: Dict[str, str] = {  # type: ignore
        "adc": "adc_mean",
        "tmp": "tmp_mean",
        "atm": "atm_mean",
        "hmd": "hmd_mean",
        "hit_top": "hit_top_mean",
        "hit_mid": "hit_mid_mean",
        "hit_btm": "hit_btm_mean",
    }
    # fmt: off
    _mean = (
        data.resample(freq)[names]
        .mean()
        .reset_index().rename(columns=renames)
    )
    # fmt: on

    # 標準偏差を求める
    renames: Dict[str, str] = {  # type: ignore
        "adc": "adc_std",
        "tmp": "tmp_std",
        "atm": "atm_std",
        "hmd": "hmd_std",
        "hit_top": "hit_top_std",
        "hit_mid": "hit_mid_std",
        "hit_btm": "hit_btm_std",
    }
    # fmt: off
    _std = (
        data.resample(freq)[names]
        .std()
        .reset_index().rename(columns=renames)
    )
    # fmt: on

    # データフレームをまとめる
    # fmt: off
    resampled = (
        _size
        .merge(_sum, on="time")
        .merge(_mean, on="time")
        .merge(_std, on="time")
    )
    # fmt: on

    # 累積和を求める
    resampled["hit_top_sum"] = resampled["hit_top"].cumsum()
    resampled["hit_mid_sum"] = resampled["hit_mid"].cumsum()
    resampled["hit_btm_sum"] = resampled["hit_btm"].cumsum()
    resampled["adc_sum"] = resampled["adc"].cumsum()
    resampled["hit_sum"] = resampled["hit"].cumsum()

    # ヒットレートを計算する
    resampled["interval"] = interval
    resampled["hit_rate"] = resampled["hit"] / interval
    resampled["hit_top_rate"] = resampled["hit_top"] / interval
    resampled["hit_mid_rate"] = resampled["hit_mid"] / interval
    resampled["hit_btm_rate"] = resampled["hit_btm"] / interval

    # 測定時間を計算する
    epoch = resampled["time"].min()
    resampled["days"] = resampled["time"] - epoch
    resampled["seconds"] = resampled["days"].dt.total_seconds()

    return resampled


@deprecated(version="0.3.0", reason="Use preprocess.preprocess_data.")
def parse_data(data: pd.DataFrame, rules: dict, interval: int) -> pd.DataFrame:
    """
    データフレームを整理する

    1. イベントカテゴリの数だけデータフレームを分割する
    2. サンプリング間隔が設定されている場合はサンプリング集計する
    3. 分割したデータフレームを1つに統合する

    Parameters
    ----------
    data : pd.DataFrame
        入力データ
    rules : dict
        イベントカテゴリ
    interval : int
        サンプリング間隔（秒）

    Returns
    -------
    pd.DataFrame
        整理したデータ
    """
    _parsed = []
    for name, rule in rules.items():
        _sliced = data.query(rule).copy()
        _sliced = resample_data(_sliced, interval)
        _sliced["hit_type"] = name
        _sliced["condition"] = rule
        _parsed.append(_sliced)

    # 1つのデータフレームにまとめる
    # 時刻（time）とイベントタイプ（hit_type）でソートする
    parsed = pd.concat(_parsed, ignore_index=True)
    parsed = parsed.sort_values(by=["time", "hit_type"])
    return parsed


@deprecated(version="0.3.0", reason="Use preprocess.raw2csv.")
def parse_files(
    fnames: List[Path], rules: Dict[str, str], interval: int
) -> pd.DataFrame:
    loaded = load_files(fnames)
    parsed = parse_data(loaded, rules, interval)
    return parsed


@deprecated(version="0.3.0", reason="Use preprocess.run2csv.")
def parser(data: RunData, rules: dict) -> pd.DataFrame:
    fnames = data.fnames

    # Load rawdata and save as raw2gz
    savef = data.raw2gz
    loaded = load_files(fnames)
    loaded["runid"] = data.runid
    loaded["name"] = data.name

    if savef:
        loaded.to_csv(savef, index=False, compression="gzip")
        info = f"Saved as {savef}"
        logger.success(info)

    # Parse rawdata and save as raw2csv
    savef = data.raw2csv
    parsed = parse_data(loaded, rules=rules, interval=data.interval)
    parsed["runid"] = data.runid
    parsed["name"] = data.name
    parsed["desc"] = data.description

    if savef:
        parsed.to_csv(savef, index=False)
        info = f"Saved as {savef}"
        logger.success(info)



#! /usr/bin/env python3

import argparse
import csv
import sys
from datetime import datetime
from pathlib import Path

import pandas as pd
import pendulum
import serial
from deprecated import deprecated
from icecream import ic
from loguru import logger
from pydantic import BaseModel
from tqdm import tqdm

from .config import Daq


class RealEvent(BaseModel):
    """実イベント"""

    timestamp: datetime = pendulum.now()
    """測定時刻。宇宙線イベントが通過した日時"""
    top: int = 0
    """topレイヤーのヒット。0 - 10の値"""
    mid: int = 0
    """midレイヤーのヒット。0 - 10の値"""
    btm: int = 0
    """btmレイヤーのヒット。0 - 10の値"""
    adc: int = 0
    """topレイヤーにヒットがあったときのADC値。0 - 1023の値"""
    tmp: float = 0.0
    """BME280で測定した気温。[degC]"""
    atm: float = 0.0
    """BME280で測定した気圧。[Pa]"""
    hmd: float = 0.0
    """BME280で測定した湿度。[%]"""


@deprecated(version="0.6.0", reason="Use typer.Typer for CLIs.")
def init_parser():
    """[deprecated] Initialize argument parser

    :deprecated: `0.6.0`. Replaced with typer.Typer for CLIs.

    """

    now = pendulum.now().strftime("%Y%m%d")

    parser = argparse.ArgumentParser(description="OSECHI DAQ")
    parser.add_argument(
        "--saved",
        default=Path(now),
        type=Path,
        help=f"path to save data. (default={now})",
    )
    parser.add_argument(
        "-q", action="count", default=0, help="decrease verbosity. (default=0)"
    )
    parser.add_argument(
        "--skip", type=int, default=10, help="skip rows to print. (default=10)"
    )

    group1 = parser.add_argument_group("FILE", description="Options for files.")
    group1.add_argument(
        "--prefix",
        default="osechi_data",
        help="set prefix of filename. (default='osechi_data')",
    )
    group1.add_argument(
        "--suffix",
        choices=[".csv", ".dat"],
        default=".csv",
        help="set suffix of filename. (default='.csv')",
    )
    group1.add_argument(
        "--max-rows",
        metavar="ROWS",
        type=int,
        default=10000,
        help="set max. rows per file. (default=10000)",
    )
    group1.add_argument(
        "--max-files",
        metavar="FILES",
        type=int,
        default=100,
        help="set max. files to create. (default=100)",
    )
    group1.add_argument(
        "--append",
        action="store_true",
        help="append data to existing file. (default=False)",
    )

    group2 = parser.add_argument_group(
        "CONNECTION", description="Options for serial connection."
    )
    group2.add_argument(
        "--device",
        default="/dev/ttyUSB0",
        help="set USB device name. (default='/dev/ttyUSB0')",
    )
    group2.add_argument(
        "--baudrate", type=int, default=115200, help="set baudrate. (default=115200)"
    )
    group2.add_argument(
        "--timeout",
        type=int,
        default=1000,
        help="set read timeout in seconds. (default=1000)",
    )

    return parser


def init_saved(daq: Daq) -> None:
    """Initialize destination directory to save data files

    保存先のディレクトリを初期化します。
    ディレクトリが存在しない場合は、作成します。
    ディレクトリが存在する場合は、``FileExistsError``を発行し、終了（``sys.exit``）します。
    ただし。DAQオブジェクトの``--append``オプションが有効な場合は、終了せずに追記を許可します。

    :Args:
    - daq(Daq): DAQ設定オブジェクト

    :Example:

    ```python
    >>> daq = Daq()
    >>> daq.load_toml(load_from)
    ```
    """

    p = Path(daq.saved)
    try:
        p.mkdir()
        msg = f"Created new directory : {p}"
        logger.info(msg)
    except FileExistsError as e:
        msg = f"Directory already exists : {p}"
        logger.warning(msg)
        if not daq.append:
            logger.error(e)
            sys.exit()
        msg = "Overwrite data to existing files."
        logger.warning(msg)

    return


def get_savef(args: Daq, n: int) -> Path:
    """Get filename to save data

    保存するファイル名を生成します。
    DAQ設定の接頭辞（``prefix``）と拡張子（``suffix``）の値を使って
    ``{prefix}_{連番:06}.{suffix}``の形式で生成します。

    :Args:
    - args(Daq): DAQ設定オブジェクト
        - prefix: ファイルの接頭辞
        - suffix: ファイルの拡張子
    - n(int): ファイル番号

    :Example:
    ```console
    osechi_data_000000.csv
    osechi_data_000001.csv
    osechi_data_000002.csv
    ```

    """
    stem = f"{args.prefix}_{n:06}"
    fname = Path(stem).with_suffix(args.suffix)
    savef = Path(args.saved, fname)

    msg = "deprecation warning: use get_savef_with_timestamp instaed."
    logger.warning(msg)

    return savef


def get_savef_with_timestamp(args: Daq, n: int) -> Path:
    """Get filename to save data with timestamp

    作成日を含んだファイル名を生成する。
    ファイル名は、DAQ設定の接頭辞（``prefix``）、ファイルを開いた時刻（``pendulum.now``）と
    拡張子（``suffix``）の値を使って生成する。

    時刻のフォーマットは、ファイル名が分かりやすいように独自フォーマットにした。

    :Args:
        - args (Daq): Daqオブジェクト
        - n (int): ファイル番号

    :Returns:
        - savef (Path): Pathオブジェクト

    :Examples:
    ```console
    20240520/osechi_data_2024-05-20T12h34m56s_000000.csv
    20240520/osechi_data_2024-05-20T13h53m24s_000001.csv
    20240520/osechi_data_2024-05-20T14h46m23s_000000.csv  // 走り直すとリセット
    20240520/osechi_data_2024-05-20T14h36m32s_000001.csv
    ```

    """
    ts = pendulum.now().format("YYYY-MM-DDTHH[h]mm[m]ss[s]")
    stem = f"{args.prefix}_{ts}_{n:06}"
    fname = Path(stem).with_suffix(args.suffix)
    savef = Path(args.saved, fname)
    return savef


@deprecated(version="0.6.0", reason="Use run.")
def daq():
    """[deprecated] Main DAQ

    :deprecated: `0.6.0`. Use `run` instead.


    """
    ic.configureOutput(prefix="[running daq] | ")

    parser = init_parser()
    args = parser.parse_args()
    init_saved(args)

    ic(args)

    # Open serial connection

    with serial.Serial(
        args.device, baudrate=args.baudrate, timeout=args.timeout
    ) as port:
        port.rts = False
        port.dtr = False

        for nfile in tqdm(range(args.max_files), desc="files"):
            savef = get_savef(args, nfile)
            ic(savef)

            with savef.open("a") as f:
                for nrow in tqdm(range(args.max_rows), leave=False, desc="rows"):
                    now = pendulum.now().to_iso8601_string()
                    data = port.readline().decode("utf-8").strip()
                    row = f"{now} {data}"
                    if args.suffix == ".csv":
                        row = (",").join(row.split())
                    f.write(row + "\n")

                    if args.q == 0 and nrow % args.skip == 0:
                        ic(nrow, row)


def open_serial_connection(daq: Daq) -> serial.Serial:
    """Open and return a serial connection.

    シリアル通信（UART）に使うポートを準備します。
    ``with``構文で使う想定です。

    通信に使うUSBポート名（``device``）、
    ボーレート（``baudrate``）、
    通信開始／書き込みのタイムアウト秒（``timeout``）は
    DAQ用の設定ファイルで変更できるようにしてあります。

    :Args:
    - daq(Daq): DAQ設定オブジェクト
        - device: USBポート名
        - baudrate: ボーレート（通信速度）[bps]
        - timeout: タイムアウト秒 [sec]

    :Returns:
    - port(serial.Serial): 通信を開始したシリアルオブジェクト

    :Example:

    ```python
    >>> with open_serial_connection(daq) as port:
    >>>     # データ測定の処理
    >>>
    ```

    """
    port = serial.Serial(
        daq.device,
        baudrate=daq.baudrate,
        timeout=daq.timeout,
        write_timeout=daq.timeout,
    )
    port.rts = False
    port.dtr = False

    # logger.debug(f"Port opened: {port}")

    return port


def write_vth(port: serial.Serial, ch: int, vth: int):
    """Write threshold to OSECHI.

    シリアル通信（UART）を使って閾値を書き込みます。
    値を書き込んだあとに、値が読み出せるかを確認します。

    読み出した値が ``dame`` となっている場合は書き込みに失敗しています。
    その場合は、もういちど設定し直すよう警告を表示します。

    ESP32のバッファサイズの制限から、閾値は2回に分割して転送しています。

    - ``val1`` は vth を右に6ビットシフト（=64で割る）して、head を足した値
    - ``val2`` は vth を左に2ビットシフト（=4をかける）して、下位8ビットを取り出した値

    詳細はこの関数のソースコードを確認してください。
    """
    val = vth
    head = 0b10000
    val1 = head + (val >> 6)
    val2 = (val << 2) & 0xFF

    logger.debug(f"Write: ch = {ch}")
    logger.debug(f"Write: val1 = {val1:b}")
    logger.debug(f"Write: val2 = {val2:b}")

    port.write(ch.to_bytes(1, "big"))
    port.write(val1.to_bytes(1, "big"))
    port.write(val2.to_bytes(1, "big"))

    read0 = port.readline().decode("utf-8", "ignore").strip()
    read1 = port.readline().decode("utf-8", "ignore").strip()
    read2 = port.readline().decode("utf-8", "ignore").strip()

    logger.debug(f"Read: ch = {read0}")
    logger.debug(f"Read: val1 = {read1}")
    logger.debug(f"Read: val2 = {read2}")

    success = False

    if read0 == "dame":
        success = False
        msg = f"Ch{ch}: Set threshold failed. Try again."
        logger.warning(msg)
    elif read0 == str(ch):
        success = True
        msg = f"Ch{ch}: Set threshold to {vth}."
        logger.success(msg)
    else:
        success = True
        msg = f"Ch{ch}: Set threshold to {vth}. Maybe."
        logger.success(msg)
    return success


def set_vth(daq: Daq, ch: int, vth: int) -> bool:
    """Set threshold

    チャンネル番号を指定し、閾値を設定します。
    チャンネル番号は1 - 3 の範囲で指定してください。
    閾値は1 - 1023 の範囲で指定してください。

    書き込みに成功すると ``success=True`` を返します。
    書き込みに失敗した場合は、警告メッセージを表示します。
    このとき、設定済みの閾値はそのままになります。

    :Args:
    - daq(Daq): DAQ設定オブジェクト
    - ch(int): チャンネル番号。1 - 3の範囲で指定してください
    - vth(int): 閾値の値。1 - 1023 の範囲で指定してください
    """

    # check values
    if ch not in range(1, 4):
        msg = f"value of ch is out of range: {ch}"
        logger.error(msg)
        sys.exit()

    if vth not in range(1, 1024):
        msg = f"value of vth is out of range: {vth}"
        logger.error(msg)
        sys.exit()

    init_saved(daq)

    try:
        with open_serial_connection(daq) as port:
            success = write_vth(port, ch, vth)
            fname = Path(daq.saved) / daq.fname_logs

            with fname.open("a", newline="") as f:
                now = pendulum.now().to_iso8601_string()
                row = [str(now), str(ch), str(vth), str(success)]
                writer = csv.writer(f)
                writer.writerow(row)
            msg = f"Saved data to {fname}."
            logger.info(msg)
        return success
    except serial.SerialException as e:
        logger.error(e)
        msg = """Could not open the port. Device name might be wrong.
        Run 'arduino-cli board list' and check the device name.
        Edit 'daq.toml' if you needed.
        """
        logger.warning(msg)
        sys.exit()
    except Exception as e:
        logger.error(e)
        msg = """
        Unaware error occurred.
        Please think if you need to handle this error.
        """
        sys.exit()


def set_vth_retry(daq: Daq, ch: int, vth: int, max_retry: int) -> bool:
    """
    Retry setting threshold.

    :Args:
    - daq (Daq): Daqオブジェクト
    - ch (int): チャンネル番号
    - vth (int): スレッショルド値
    - max_retry (int): 設定に失敗ししたときにトライする最大回数

    :Returns:
    - success (bool): スレッショルド値の設定の結果
    """

    for i in range(max_retry):
        success = set_vth(daq, ch, vth)
        if success:
            return True
        msg = f"Retry: {i} / {max_retry} times."
        logger.warning(msg)

    return False


def read_serial_data(port: serial.Serial) -> list:
    """Read serial data from port.

    OSECHIが接続されているポートからデータを読み出します。
    引数にしていするポートは、あらかじめ開いたものを渡してください。
    ``run_daq``や``time_daq``でデータを取得するために使います。

    :Args:
    - port (serial.Serial): Serialオブジェクト

    :Returns:
    - row (list): 読み出した時刻を追加したデータのリスト

    :Examples:
    ```python
    >>> with open_serial_connection() as port:
    >>>     row = read_serialc_data(port)
    >>>     row
    [日付, top, mid, btm, adc, tmp, atm, hmd]
    ```

    :TODO:
    - list -> RealEventオブジェクトに変換する
    - タイムアウトしたときの処理を整理する

    """
    now = pendulum.now().to_iso8601_string()
    data = port.readline().decode("UTF-8").strip()
    if len(data) == 0:
        msg = f"No data to readout. Timed-out: {port.timeout}"
        logger.warning(msg)
    row = f"{now} {data}".split()
    return row


def save_serial_data(f, daq: Daq, port: serial.Serial) -> list:
    """
    :Args:
    - f: ファイルポインタ
    - daq (Daq): Daqオブジェクト
    - port (serial.Serial): Serialオブジェクト

    :Return:
    - rows (list[list]): 取得したデータのリスト

    :TODO:
    - Daqオブジェクトに依存しない関数にしたい（ジェネレーターにするのがいいのかな？）
    - pd.DataFrameを返した方がいいかもしれない？
    """
    max_rows = daq.max_rows
    rows = []
    for nrow in tqdm(range(max_rows), leave=False, desc="rows"):
        row = read_serial_data(port)
        rows.append(row)
        if daq.suffix == ".csv":
            writer = csv.writer(f)
            writer.writerow(row)
        else:
            writer = csv.writer(f, delimiter=" ")
            writer.writerow(row)
        f.flush()
    return rows


def run_daq(daq: Daq) -> None:
    """メインDAQ

    OSECHIが接続されたUSBポートとシリアル通信をして、データ取得する。
    指定したファイル数と行数をでループ処理する。

    :TODO:
    - 関数名をより適切なものに変えたい
    - while_daq みたいなのを作ってもいいかもしれない
    """
    # Open serial connection
    with open_serial_connection(daq) as port:
        init_saved(daq)

        for nfile in tqdm(range(daq.max_files), desc="files"):
            savef = get_savef_with_timestamp(daq, nfile)
            msg = f"Saving data to: {savef}."
            logger.info(msg)
            logger.info("Press Ctrl-c to stop.")

            with savef.open("a") as f:
                save_serial_data(f, daq, port)

            msg = f"Saved data to: {savef}."
            logger.success(msg)


def run(args: Daq):
    """メインのDAQ

    run_daqのラッパー。例外処理などで囲んだもの。

    :versionadded: `0.6.0`.
    """

    try:
        run_daq(args)
    except serial.SerialException as e:
        logger.error(e)
        msg = """Could not open the port. Device name might be wrong.
        Run 'arduino-cli board list' and check the device name.
        Edit 'daq.toml' if you needed.
        """
        logger.warning(msg)
        sys.exit()
    except KeyboardInterrupt as e:
        logger.warning(e)
        msg = """Quit."""
        logger.info(msg)
        sys.exit()
    except Exception as e:
        logger.error(e)
        msg = """Exit.
        Unaware error occurred.
        Please think if you need to handle this error.
        """
        logger.error(msg)
        sys.exit()


def time_daq(args: Daq, duration: int) -> pd.DataFrame:
    """
    測定時間を指定してDAQを走らせます。

    :Args:
    - args (Daq): Daqオブジェクト
    - duration (int): 測定時間を秒で指定

    :Returns:
    - data (pd.DataFrame): 測定結果のデータフレーム
    """

    rows = []
    with open_serial_connection(args) as port:
        init_saved(args)

        logger.debug("Port opened.")
        daq_start = pendulum.now()
        daq_stop = daq_start.add(seconds=duration)

        logger.debug(f"- DAQ Started: {daq_start}")
        logger.debug(f"- DAQ Stop   : {daq_stop}")

        while pendulum.now() < daq_stop:
            row = read_serial_data(port)
            event = RealEvent()
            event.timestamp = pendulum.parse(row[0])
            event.top = int(row[1])
            event.mid = int(row[2])
            event.btm = int(row[3])
            event.adc = int(row[4])
            event.tmp = float(row[5])
            event.atm = float(row[6])
            event.hmd = float(row[7])
            rows.append(event.model_dump())
    daq_end = pendulum.now()
    diff = daq_end - daq_start
    elapsed_time = diff.in_seconds()
    logger.debug(f"- DAQ Closed : {daq_end} / Elapsed: {elapsed_time} sec.")
    data = pd.DataFrame(rows)
    return data


@deprecated(
    version="0.14.0",
    reason="Will be deprecated. Use threshold.scan_threshold_by_channel instead.",
)
def scan_ch_vth(daq: Daq, duration: int, ch: int, vth: int) -> list:
    """
    Run threshold scan.

    :Args:
    - daq (Daq): Daqオブジェクト
    - duration (int): 測定時間（秒）
    - ch (int): 測定するチャンネル番号
    - vth (int): スレッショルド値

    :Returns:
    - data (list): [測定時刻、チャンネル番号、スレッショルド値、イベント数]のリストを返します。
    """

    logger.warning("Will be deprecated. Please use threshold.scan_by_channel instead.")

    # Try to set the threshold
    if not set_vth_retry(daq, ch, vth, 3):
        msg = f"Failed to set threshold: ch{ch} - {vth}"
        logger.error(msg)
        return []

    # Collect data
    try:
        rows = time_daq(daq, duration)
        counts = len(rows)
        tmp = rows["tmp"].mean()
        atm = rows["atm"].mean()
        hmd = rows["hmd"].mean()
        fname = get_savef_with_timestamp(daq, ch)
        rows.to_csv(fname, index=False)
        msg = f"Saved data to: {fname}"
        logger.info(msg)
    except Exception as e:
        msg = f"Failed to collect data due to: {str(e)}"
        logger.error(msg)
        counts = 0
        tmp = 0
        atm = 0
        hmd = 0

    # Save Summary
    now = pendulum.now().to_iso8601_string()
    data = [now, duration, ch, vth, counts, tmp, atm, hmd]
    fname = Path(daq.saved) / daq.fname_scan
    with fname.open("a", newline="") as f:
        writer = csv.writer(f)
        writer.writerow(data)
    msg = f"Added data to: {fname}"
    logger.info(msg)

    return data


@deprecated(
    version="0.14.0",
    reason="Will be deprecated. Use threshold.scan_thresholds instead.",
)
def scan_ch_thresholds(
    daq: Daq, duration: int, ch: int, thresholds: list[int]
) -> list[list]:
    """
    Run threshold scan.

    :Args:
    - daq (Daq): Daqオブジェクト
    - duration (int): 測定時間（秒）
    - ch (int): チャンネル番号
    - thresholds (list[int]): スレッショルド値のリスト

    :Returns:
    - rows (list[list]): [測定時刻、チャンネル番号、スレッショルド値、イベント数]のリスト
    """

    logger.warning("Will be deprecated. Please use threshold.scan_thresholds instead.")
    # Estimated time for scan
    msg = f"Number of points: {len(thresholds)}"
    logger.info(msg)
    estimated_time = len(thresholds) * duration
    msg = f"Estimated time: {estimated_time} sec."
    logger.info(msg)

    # すべてのチャンネルの閾値を高くする
    set_vth_retry(daq, 1, 500, 5)
    set_vth_retry(daq, 2, 500, 5)
    set_vth_retry(daq, 3, 500, 5)

    rows = []
    n = len(thresholds)
    for i, vth in enumerate(thresholds):
        msg = "-" * 40 + f"[{i+1:2d}/{n:2d}: {vth}]"
        logger.info(msg)
        row = scan_ch_vth(daq, duration, ch, vth)
        if row:
            rows.append(row)

    return rows


if __name__ == "__main__":
    try:
        daq()
    except KeyboardInterrupt as e:
        ic(e)
        sys.exit()



import sys
from pathlib import Path

import pandas as pd
import pendulum

# from icecream import ic
from loguru import logger


def reformat_datetime(datetime: str) -> str:
    """
    日時を変換する

    Example
    -------

    - ``2022-05-18T20:48:14`` -> ``2022-05-18T20:48:14``
    - ``2022-06-02T18:52:36.442190+09:00`` -> ``2022-06-02T18:52:36``

    """
    dt = pendulum.parse(datetime)
    fmt = dt.format("YYYY-MM-DDTHH:mm:ss")
    return fmt


def load_raw_data(fname: Path, **kwargs) -> pd.DataFrame:
    """
    DAQで取得したデータ形式を ``pd.DataFrame`` に変換して読み込む

    Parameters
    ----------
    fname : Path
        生データのファイル名

    kwargs
        ``pd.read_csv`` の引数を想定

    Returns
    -------
    pd.DataFrame
        データフレーム

    Notes
    -----
    - ファイルの保存形式（＝拡張子）で ``pd.read_csv`` の引数をちょっと変える必要がある。拡張子は ``.dat`` / ``.csv`` だけOKにしてあり、それ以外の場合は ``sys.exit`` する。
    - ``.csv`` はそのままカンマ区切り、``.dat`` はスペース区切り（ ``sep=" "`` ）として、 ``pd.read_csv`` する。
    - ファイル内のカラム名は適当だったり、なかったりする。適切なカラム名を付与する。カラム名はこの関数内にハードコードしている。
    - イベント時刻（ ``time`` ）は ``pd.datetime`` オブジェクトに変換する。使ったDAQのバージョンによって記録された日時の形式が異なるので、それを内部で変換している。
    - 各レイヤーのヒットの有無（ ``True`` / ``False`` ）を確認して、ヒット用のカラム（ ``hit_top`` / ``hit_mid`` / ``hit_btm`` ）に保存する。現在、各レイヤーの値自体には意味がない。そのうち光量など意味を持たせる可能性はあるかも？
    - ヒットのあったレイヤーのパターンを計算して 8ビット で表現する。``hit_type = hit_top * 4 + hit_mid * 2 + hit_btm * 1``
    """

    _names = ["time", "top", "mid", "btm", "adc", "tmp", "atm", "hmd"]
    suffix = fname.suffix
    if suffix == ".dat":
        data = pd.read_csv(fname, names=_names, sep=" ", comment="t", **kwargs)
    elif suffix == ".csv":
        data = pd.read_csv(fname, names=_names, **kwargs)
    else:
        error = f"Unknown suffix : {suffix}"
        logger.error(error)
        sys.exit()

    data["time"] = data["time"].apply(reformat_datetime)
    data["time"] = pd.to_datetime(data["time"], format="ISO8601")
    # 各レイヤーのヒットの有無を確認する
    data["hit_top"] = data["top"] > 0
    data["hit_mid"] = data["mid"] > 0
    data["hit_btm"] = data["btm"] > 0
    # ヒットのあったレイヤーのパターンを8ビットで表現する
    data["hit_type"] = data["hit_top"] * 4 + data["hit_mid"] * 2 + data["hit_btm"] * 1
    return data


def load_files(fnames: list[Path]) -> pd.DataFrame:
    """複数のファイルを読み込み、単一のデータフレームに変換する

    Parameters
    ----------
    fnames : list[Path]
        読み込むファイルの一覧

    Returns
    -------
    pd.DataFrame
        ファイルごとに変換したデータフレームをすべて結合したデータフレーム

    Notes
    -----
    - ``fnames`` で列挙されたファイルごとに、データフレームに変換する
    - ファイル名の一覧が空の場合は終了する
    - 個々のデータフレームを結合して単一のデータフレームを作成する
    - 結合したデータフレームは、時刻（``time``）でソートしておく
    """
    if len(fnames) == 0:
        error = "No files listed"
        logger.error(error)
        sys.exit()

    _loaded = []
    for fname in fnames:
        _data = load_raw_data(fname)
        _loaded.append(_data)
    loaded = pd.concat(_loaded, ignore_index=True)
    loaded = loaded.sort_values(["time"])
    debug = f"Entries : {len(loaded)}"
    logger.debug(debug)
    return loaded
